#include "xdvi-config.h"
#include "xdvi.h"
#ifdef	KANJI

#include "jfm.h"
#include "util.h"

#ifndef	SEEK_SET
#define	SEEK_SET	0
#endif

/**** read_jfm0: borrowed from jxdvi-NEWS ****/
static int read_jfm0(FILE *fp, struct jfm *j)
{
	int i;

	fseek(fp, 0L, SEEK_SET);

	/* read jfm table field. */
	for (i = 0; i <= J_NP; i++)
	    j->table[i] = (short) get_bytes(fp, 2);

	/* read jfm header */
	j->check_sum = get_bytes(fp, 4);
	j->design_size = get_bytes(fp, 4);
	fseek(fp, (long) (4 * (j->table[J_LH] - 2)), 1);
    
	/* read jfm char_type */
	j->type = xmalloc(sizeof(struct jfm_char_type) * j->table[J_NT]);
	for (i=0; i < j->table[J_NT]; i++) {
	    j->type[i].code = (short) get_bytes(fp, 2);
	    j->type[i].index = (short) get_bytes(fp, 2);
	}
    
	/* read jfm char_info */
	j->info = xmalloc(sizeof(struct jfm_char_info) * (j->table[J_EC] + 1));
	for (i = 0; i <= j->table[J_EC]; i++) {
	    j->info[i].width_ix = (unsigned char) get_byte(fp);
	    j->info[i].height_depth_ix = (unsigned char) get_byte(fp);
	    j->info[i].italic_ix_tag = (unsigned char) get_byte(fp);
	    j->info[i].remainder_ix = (unsigned char) get_byte(fp);
	}
    
	/* read jfm width */
	j->width = xmalloc(sizeof(unsigned long) * j->table[J_NW]);
	for (i = 0; i < j->table[J_NW]; i++)
	    j->width[i] = get_bytes(fp, 4);

	/* read jfm height */
	j->height = xmalloc(sizeof(unsigned long) * j->table[J_NH]);
	for (i = 0; i < j->table[J_NH]; i++)
	    j->height[i] = get_bytes(fp, 4);

	/* read jfm depth */
	j->depth = xmalloc(sizeof(unsigned long) * j->table[J_ND]);
	for (i = 0; i < j->table[J_ND]; i++)
	    j->depth[i] = get_bytes(fp, 4);

	/* read jfm italic */
	j->italic = xmalloc(sizeof(unsigned long) * j->table[J_NI]);
	for (i = 0; i < j->table[J_NI]; i++)
	    j->italic[i] = get_bytes(fp, 4);
	return 0;
}

struct jfm *read_jfm(FILE *fp, char *fontname)
{
	struct jfmchain {
	    struct jfmchain *next;
	    struct jfm *jfm;
	};
	static struct jfmchain *jfmtop = NULL;
	struct jfmchain *jcp, *new;

	for (jcp = jfmtop; jcp; jcp = jcp->next)
	    if (0 == strcmp(fontname, jcp->jfm->fontname))
		return jcp->jfm;

	new = xmalloc(sizeof(struct jfmchain));
	new->next = NULL;
	new->jfm = xmalloc(sizeof(struct jfm));
	if (read_jfm0(fp, new->jfm)) {
	    fprintf(stderr, "Cannot open metric file: %s\n", fontname);
	    free(new->jfm);
	    free(new);
	    return NULL;
	}
	new->jfm->fontname = xstrdup(fontname);
	if (NULL == jfmtop) {
	    jfmtop = new;
	} else {
	    for (jcp = jfmtop; jcp->next; jcp = jcp->next)
		;
	    jcp->next = new;
	}

	return new->jfm;
}
#endif	/* KANJI */
