#include "xdvi-config.h"
#include "xdvi.h"

#ifdef	USE_ZEIT
#include <ft2build.h>
#include FT_FREETYPE_H
#include <stdio.h>

#include "kpathsea/tex-file.h"
#include "vf2ft.h"
#include "dvi-init.h"
#include "util.h"
#include "vfontmap.h"
#include "ft2vert.h"
#include "jisx0208.h"

static struct jisv_st {
  int low, hi;
} jisv[] = {
/* converted from CMap 'V' */
/* sed -e '/^</!d' -e 's|<\(....\)> <\(....\)>.*|  { 0x\1, 0x\2 },|' < V */
  { 0x2122, 0x2123 },
  { 0x2131, 0x2132 },
  { 0x213c, 0x213e },
  { 0x2141, 0x2145 },
  { 0x214a, 0x215b },
  { 0x2161, 0x2161 },
  { 0x2421, 0x2421 },
  { 0x2423, 0x2423 },
  { 0x2425, 0x2425 },
  { 0x2427, 0x2427 },
  { 0x2429, 0x2429 },
  { 0x2443, 0x2443 },
  { 0x2463, 0x2463 },
  { 0x2465, 0x2465 },
  { 0x2467, 0x2467 },
  { 0x246e, 0x246e },
  { 0x2521, 0x2521 },
  { 0x2523, 0x2523 },
  { 0x2525, 0x2525 },
  { 0x2527, 0x2527 },
  { 0x2529, 0x2529 },
  { 0x2543, 0x2543 },
  { 0x2563, 0x2563 },
  { 0x2565, 0x2565 },
  { 0x2567, 0x2567 },
  { 0x256e, 0x256e },
  { 0x2575, 0x2576 },
};

Boolean is_jis_vertical(int jis)
{
    size_t i;

    for (i=0; i<sizeof(jisv)/sizeof(*jisv); i++) {
	if (jisv[i].low <= jis && jis <= jisv[i].hi) return True;
    }
    return False;
}

static const char *get_font_pathname(char *name)
{
    char *s = NULL;
    FILE *fp;

    if ((fp=fopen(name, "r")) != NULL) {
	fclose(fp);
	s = name;
    }
    if (s == NULL) s = kpse_find_file(name, kpse_opentype_format, true);
    if (s == NULL) s = kpse_find_file(name, kpse_truetype_format, true);
    if (s == NULL) s = kpse_find_file(name, kpse_miscfonts_format, true);
    if (s == NULL) s = kpse_find_file(name, kpse_program_text_format, true);
    if (s == NULL) s = name;
    return s;
}

#define  FT_FONTS_MAX  100
static FT_Library  library;
static int ft_error;

static struct ft_font {
  FT_Face face;
  short   height;
  short   width;
  short   ascend;
  int     encoding;
  struct ft2vert_st *ft2vert;
} *ft_fonts[FT_FONTS_MAX];

static int ft_fonts_max = 0;

int VF_Init(char *vfcap)
{
  UNUSED(vfcap);
  ft_error = FT_Init_FreeType( &library );
  if ( ft_error ) {
    fprintf(stderr, "FT2: Initialization Error\n");
    exit(100);
  }
  return 0;
}

int VF_OpenFont(char *fontname, struct vfontmap *map)
{
  const char *fontpath = get_font_pathname(fontname);
  struct ft_font *font;

  if( ft_fonts_max == FT_FONTS_MAX ) {
    fprintf(stderr, "Fonts exceed limit\n");
    return -1;
  }

  font = xmalloc(sizeof(*font));
  ft_fonts[ft_fonts_max] = font;

  ft_error = FT_New_Face( library, fontpath, map->face_index, &font->face );
  if (ft_error != 0) {
    fprintf(stderr, "FT2: Open Font Error (%s).  Error code = %d \n",
	    fontpath, ft_error);
    ft_error = 0;
    font->face = NULL;
  } else {
    font->height   = 0;
    font->width    = 0;
    font->encoding = map->encoding;
    if (map->encoding == ENC_JIS_V || map->encoding == ENC_UNICODE_V) {
	font->ft2vert = ft2vert_init(font->face);
    } else {
	font->ft2vert = NULL;
    }
    if (map->extend != 1.0 || map->slant != 0.0) {
	/* t1_modify_fm() in tetex-src-3.0/texk/dvipsk/writet1.c */
	FT_Matrix matrix;
	matrix.xx = (int)(map->extend * (1<<16));
	matrix.xy = (int)(map->extend * map->slant * (1<<16));
	matrix.yx = 0;
	matrix.yy = 1<<16;
	FT_Set_Transform( font->face, &matrix, 0 );
	/* TODO: vertical font */
    }
  }
  return ft_fonts_max++;
}

int VF_GetBitmap(int char_code, int fid, int w, int h,
		 int bw, int bo, char *bm_buf)
{
  int  x, xmax, xshift, y, ymax, yoffset;
  unsigned char *src, *dst;
  FT_UInt glyph_index;
  struct ft_font *font = ft_fonts[fid];
  FT_GlyphSlot glyph;
  UNUSED(bo);

  memset(bm_buf, 0, bw * h);
  if (font->face == NULL) return 0;
  glyph = font->face->glyph;
  if (font->width != w ||
      font->height != h) {
    FT_Set_Pixel_Sizes( font->face, w, h );
    font->height = h;
    font->width  = w;
    font->ascend =
      h* font->face->ascender
      / (font->face->ascender - font->face->descender);
  }

  switch (font->encoding) {
  case ENC_IDENTITY:
    glyph_index = char_code;
    break;
  case ENC_UNICODE_H: case ENC_UNICODE_V:
    glyph_index = FT_Get_Char_Index( font->face, char_code );
    break;
  case ENC_JIS_H: case ENC_JIS_V:
    glyph_index = FT_Get_Char_Index( font->face, jis2uni(char_code) );
    break;
  default:
    glyph_index = 0;
    fprintf(stderr, "VF_GetBitmap: internal error\n");
    break;
  }
  if ((font->encoding == ENC_JIS_V && is_jis_vertical(char_code)) ||
      font->encoding == ENC_UNICODE_V) {
      glyph_index = ft2vert_get_gid( font->ft2vert, glyph_index );
  }
  FT_Load_Glyph( font->face, glyph_index, FT_LOAD_DEFAULT );
  FT_Render_Glyph( glyph, FT_RENDER_MODE_MONO );

    dst  = bm_buf;
    src  = glyph->bitmap.buffer;
    xmax = glyph->bitmap.pitch;
    ymax = glyph->bitmap.rows;

    if (glyph->bitmap_left >= 0) {  // to right
	dst   += glyph->bitmap_left / 8;
	xshift = glyph->bitmap_left % 8;
    } else { // to left
	src   +=   (-glyph->bitmap_left) / 8;
	xshift = -((-glyph->bitmap_left) % 8);
    }
    yoffset = font->ascend - glyph->bitmap_top;
    if (yoffset >= 0) { // to down
	dst += yoffset * bw;
	if (ymax > h - yoffset) ymax = h - yoffset;
    } else { // to up
	src  += (-yoffset) * xmax;
	ymax -= (-yoffset);
    }

    for ( y=0; y<ymax; y++ ) {
	for ( x=0; x<xmax; x++ ) {
	    if ( xshift >= 0 ) {
		*dst++ |= *src   >> xshift;
		*dst   |= *src++ << (8 - xshift);
	    } else {
		*dst   |= *src++ << -xshift;
		*dst++ |= *src   >> (8 + xshift);
	    }
	}
	dst += bw - xmax;
    }
    return font->ascend;
}

#endif	/* USE_ZEIT */
